///<reference path="../../../.d.ts"/>
"use strict";
var net = require("net");
var ref = require("ref");
var os = require("os");
var path = require("path");
var util = require("util");
var iosCore = require("./ios-core");
var iOSProxyServices = require("./ios-proxy-services");
var CoreTypes = iosCore.CoreTypes;
var IOSDevice = (function () {
    function IOSDevice(devicePointer, $childProcess, $coreFoundation, $errors, $fs, $injector, $logger, $mobileDevice, $staticConfig, $devicePlatformsConstants, $hostInfo, $options) {
        this.devicePointer = devicePointer;
        this.$childProcess = $childProcess;
        this.$coreFoundation = $coreFoundation;
        this.$errors = $errors;
        this.$fs = $fs;
        this.$injector = $injector;
        this.$logger = $logger;
        this.$mobileDevice = $mobileDevice;
        this.$staticConfig = $staticConfig;
        this.$devicePlatformsConstants = $devicePlatformsConstants;
        this.$hostInfo = $hostInfo;
        this.$options = $options;
        this.identifier = null;
        this.voidPtr = ref.refType(ref.types.void);
        this.mountImageCallbackPtr = null;
        this.uninstallApplicationCallbackPtr = null;
        this.mountImageCallbackPtr = CoreTypes.am_device_mount_image_callback.toPointer(IOSDevice.mountImageCallback);
        this.uninstallApplicationCallbackPtr = CoreTypes.am_device_mount_image_callback.toPointer(IOSDevice.uninstallCallback);
    }
    IOSDevice.mountImageCallback = function (dictionary, user) {
        var coreFoundation = $injector.resolve("coreFoundation");
        var logger = $injector.resolve("logger");
        var jsDictionary = coreFoundation.cfTypeTo(dictionary);
        logger.info("[Mounting] %s", jsDictionary["Status"]);
    };
    IOSDevice.uninstallCallback = function (dictionary, user) {
        IOSDevice.showStatus("Uninstalling", dictionary);
    };
    IOSDevice.showStatus = function (action, dictionary) {
        var coreFoundation = $injector.resolve("coreFoundation");
        var logger = $injector.resolve("logger");
        var jsDictionary = coreFoundation.cfTypeTo(dictionary);
        var output = [util.format("[%s]", action)];
        var append = function (value) {
            if (jsDictionary[value]) {
                output.push(util.format("%s: %s", value, jsDictionary[value]));
            }
        };
        if (jsDictionary) {
            append("PercentComplete");
            append("Status");
            append("Path");
        }
        logger.info(output.join(" "));
    };
    IOSDevice.prototype.getPlatform = function () {
        return this.$devicePlatformsConstants.iOS;
    };
    IOSDevice.prototype.getIdentifier = function () {
        if (this.identifier == null) {
            this.identifier = this.$coreFoundation.convertCFStringToCString(this.$mobileDevice.deviceCopyDeviceIdentifier(this.devicePointer));
        }
        return this.identifier;
    };
    IOSDevice.prototype.getDisplayName = function () {
        return this.getValue("ProductType");
    };
    IOSDevice.prototype.getModel = function () {
        return this.getValue("ProductType");
    };
    IOSDevice.prototype.getVersion = function () {
        return this.getValue("ProductVersion");
    };
    IOSDevice.prototype.getVendor = function () {
        return "Apple";
    };
    IOSDevice.prototype.getValue = function (value) {
        this.connect();
        this.startSession();
        try {
            var cfValue = this.$coreFoundation.createCFString(value);
            return this.$coreFoundation.convertCFStringToCString(this.$mobileDevice.deviceCopyValue(this.devicePointer, null, cfValue));
        }
        finally {
            this.stopSession();
            this.disconnect();
        }
    };
    IOSDevice.prototype.validateResult = function (result, error) {
        if (result !== 0) {
            this.$errors.fail(util.format("%s. Result code is: %s", error, result));
        }
    };
    IOSDevice.prototype.isPaired = function () {
        return this.$mobileDevice.deviceIsPaired(this.devicePointer) != 0;
    };
    IOSDevice.prototype.pair = function () {
        var result = this.$mobileDevice.devicePair(this.devicePointer);
        this.validateResult(result, "If your phone is locked with a passcode, unlock then reconnect it");
        return result;
    };
    IOSDevice.prototype.validatePairing = function () {
        var result = this.$mobileDevice.deviceValidatePairing(this.devicePointer);
        this.validateResult(result, "Unable to validate pairing");
        return result;
    };
    IOSDevice.prototype.connect = function () {
        var result = this.$mobileDevice.deviceConnect(this.devicePointer);
        this.validateResult(result, "Unable to connect to device");
        if (!this.isPaired()) {
            this.pair();
        }
        return this.validatePairing();
    };
    IOSDevice.prototype.disconnect = function () {
        var result = this.$mobileDevice.deviceDisconnect(this.devicePointer);
        this.validateResult(result, "Unable to disconnect from device");
    };
    IOSDevice.prototype.startSession = function () {
        var result = this.$mobileDevice.deviceStartSession(this.devicePointer);
        this.validateResult(result, "Unable to start session");
    };
    IOSDevice.prototype.stopSession = function () {
        var result = this.$mobileDevice.deviceStopSession(this.devicePointer);
        this.validateResult(result, "Unable to stop session");
    };
    IOSDevice.prototype.getDeviceValue = function (value) {
        var deviceCopyValue = this.$mobileDevice.deviceCopyValue(this.devicePointer, null, this.$coreFoundation.createCFString(value));
        return this.$coreFoundation.convertCFStringToCString(deviceCopyValue);
    };
    IOSDevice.prototype.lookupApplications = function () {
        var _this = this;
        var func = function () {
            var dictionaryPointer = ref.alloc(CoreTypes.cfDictionaryRef);
            var result = _this.$mobileDevice.deviceLookupApplications(_this.devicePointer, 0, dictionaryPointer);
            if (result !== 0) {
                _this.$errors.fail("Invalid result code %s from device lookup applications.", result);
            }
            var cfDictionary = dictionaryPointer.deref();
            var jsDictionary = _this.$coreFoundation.cfTypeTo(cfDictionary);
            return jsDictionary;
        };
        return this.tryToExecuteFunction(func);
    };
    IOSDevice.prototype.findDeveloperDirectory = function () {
        var _this = this;
        return (function () {
            var childProcess = _this.$childProcess.spawnFromEvent("xcode-select", ["-print-path"], "close").wait();
            return childProcess.stdout.trim();
        }).future()();
    };
    IOSDevice.prototype.tryToExecuteFunction = function (func) {
        this.connect();
        try {
            this.startSession();
            try {
                return func.apply(this, []);
            }
            finally {
                this.stopSession();
            }
        }
        finally {
            this.disconnect();
        }
    };
    IOSDevice.prototype.findDeveloperDiskImageDirectoryPath = function () {
        var _this = this;
        return (function () {
            var developerDirectory = _this.findDeveloperDirectory().wait();
            var buildVersion = _this.getDeviceValue("BuildVersion");
            var productVersion = _this.getDeviceValue("ProductVersion");
            var productVersionParts = productVersion.split(".");
            var productMajorVersion = productVersionParts[0];
            var productMinorVersion = productVersionParts[1];
            var developerDiskImagePath = path.join(developerDirectory, "Platforms", "iPhoneOS.platform", "DeviceSupport");
            var supportPaths = _this.$fs.readDirectory(developerDiskImagePath).wait();
            var supportPath = null;
            _.each(supportPaths, function (sp) {
                var parts = sp.split(' ');
                var version = parts[0];
                var versionParts = version.split(".");
                var supportPathData = {
                    version: version,
                    majorVersion: versionParts[0],
                    minorVersion: versionParts[1],
                    build: parts.length > 1 ? parts[1].replace(/[()]/, function () { return ""; }) : null,
                    path: path.join(developerDiskImagePath, sp)
                };
                if (supportPathData.majorVersion === productMajorVersion) {
                    if (!supportPath) {
                        supportPath = supportPathData;
                    }
                    else {
                        if (supportPathData.minorVersion === productMinorVersion) {
                            if (supportPathData.build === buildVersion) {
                                supportPath = supportPathData;
                            }
                            else {
                                if (supportPath.build !== supportPathData.build || supportPath.build === null) {
                                    supportPath = supportPathData;
                                }
                            }
                        }
                    }
                }
            });
            if (!supportPath) {
                _this.$errors.fail("Unable to find device support path.");
            }
            return supportPath.path;
        }).future()();
    };
    IOSDevice.prototype.mountImage = function () {
        var _this = this;
        return (function () {
            var imagePath = _this.$options.ddi;
            if (_this.$hostInfo.isWindows) {
                if (!imagePath) {
                    _this.$errors.fail("On windows operating system you must specify the path to developer disk image using --ddi option");
                }
                var imageSignature = _this.$fs.readFile(util.format("%s.signature", imagePath)).wait();
                var imageSize = _this.$fs.getFsStats(imagePath).wait().size;
                var imageMounterService = _this.startService(iOSProxyServices.MobileServices.MOBILE_IMAGE_MOUNTER);
                var plistService = _this.$injector.resolve(iosCore.PlistService, { service: imageMounterService, format: CoreTypes.kCFPropertyListXMLFormat_v1_0 });
                var result = plistService.exchange({
                    Command: "ReceiveBytes",
                    ImageSize: imageSize,
                    ImageType: "Developer",
                    ImageSignature: imageSignature
                }).wait();
                if (result.Status === "ReceiveBytesAck") {
                    var fileData = _this.$fs.readFile(imagePath).wait();
                    plistService.sendAll(fileData);
                }
                else {
                    var afcService = _this.startService(iOSProxyServices.MobileServices.APPLE_FILE_CONNECTION);
                    var afcClient = _this.$injector.resolve(iOSProxyServices.AfcClient, { service: afcService });
                    afcClient.transfer(imagePath, "PublicStaging/staging.dimage").wait();
                }
                try {
                    result = plistService.exchange({
                        Command: "MountImage",
                        ImageType: "Developer",
                        ImageSignature: imageSignature,
                        ImagePath: "/let/mobile/Media/PublicStaging/staging.dimage"
                    }).wait();
                    if (result.Error) {
                        _this.$errors.fail("Unable to mount image. %s", result.Error);
                    }
                    if (result.Status) {
                        _this.$logger.info("Mount image: %s", result.Status);
                    }
                }
                finally {
                    plistService.close();
                }
            }
            else {
                var func = function () {
                    var developerDiskImageDirectoryPath = _this.findDeveloperDiskImageDirectoryPath().wait();
                    var imagePath = path.join(developerDiskImageDirectoryPath, "DeveloperDiskImage.dmg");
                    _this.$logger.info("Mounting %s", imagePath);
                    var signature = _this.$fs.readFile(util.format("%s.signature", imagePath)).wait();
                    var cfImagePath = _this.$coreFoundation.createCFString(imagePath);
                    var cfOptions = _this.$coreFoundation.cfTypeFrom({
                        ImageType: "Developer",
                        ImageSignature: signature
                    });
                    var result = _this.$mobileDevice.deviceMountImage(_this.devicePointer, cfImagePath, cfOptions, _this.mountImageCallbackPtr);
                    if (result !== 0 && result !== IOSDevice.IMAGE_ALREADY_MOUNTED_ERROR_CODE) {
                        if (result === IOSDevice.INCOMPATIBLE_IMAGE_SIGNATURE_ERROR_CODE) {
                            _this.$logger.warn("Unable to mount image %s on device %s.", imagePath, _this.identifier);
                        }
                        else {
                            _this.$errors.fail("Unable to mount image on device.");
                        }
                    }
                };
                _this.tryToExecuteFunction(func);
            }
        }).future()();
    };
    IOSDevice.prototype.getInterfaceType = function () {
        return this.$mobileDevice.deviceGetInterfaceType(this.devicePointer);
    };
    IOSDevice.prototype.startHouseArrestService = function (bundleId) {
        var _this = this;
        var func = function () {
            var fdRef = ref.alloc("int");
            var result = _this.$mobileDevice.deviceStartHouseArrestService(_this.devicePointer, _this.$coreFoundation.createCFString(bundleId), null, fdRef);
            var fd = fdRef.deref();
            if (result !== 0) {
                _this.$errors.fail("AMDeviceStartHouseArrestService returned %s", result);
            }
            return fd;
        };
        return this.tryToExecuteFunction(func);
    };
    IOSDevice.prototype.startService = function (serviceName) {
        var _this = this;
        var func = function () {
            var socket = ref.alloc("int");
            var result = _this.$mobileDevice.deviceStartService(_this.devicePointer, _this.$coreFoundation.createCFString(serviceName), socket);
            _this.validateResult(result, "Unable to start service");
            return ref.deref(socket);
        };
        return this.tryToExecuteFunction(func);
    };
    IOSDevice.prototype.deploy = function (packageFile, packageName) {
        var _this = this;
        return (function () {
            var installationProxy = _this.$injector.resolve(iOSProxyServices.InstallationProxyClient, { device: _this });
            installationProxy.deployApplication(packageFile).wait();
            installationProxy.closeSocket();
        }).future()();
    };
    IOSDevice.prototype.sync = function (localToDevicePaths, appIdentifier, liveSyncUrl, options) {
        var _this = this;
        if (options === void 0) { options = {}; }
        return (function () {
            var houseArrestClient = _this.$injector.resolve(iOSProxyServices.HouseArrestClient, { device: _this });
            var afcClientForContainer = houseArrestClient.getAfcClientForAppContainer(appIdentifier.appIdentifier);
            afcClientForContainer.transferCollection(localToDevicePaths).wait();
            houseArrestClient.closeSocket();
            if (!_this.$options.skipRefresh) {
                var afcClientForContainer_1 = houseArrestClient.getAfcClientForAppContainer(appIdentifier.appIdentifier);
                afcClientForContainer_1.deleteFile("/Library/Preferences/ServerInfo.plist");
                houseArrestClient.closeSocket();
                var notificationProxyClient = _this.$injector.resolve(iOSProxyServices.NotificationProxyClient, { device: _this });
                notificationProxyClient.postNotification("com.telerik.app.refreshWebView");
                notificationProxyClient.closeSocket();
            }
            _this.$logger.out("Successfully synced device with identifier '%s'", _this.getIdentifier());
        }).future()();
    };
    IOSDevice.prototype.openDeviceLogStream = function () {
        var iOSSystemLog = this.$injector.resolve(iOSProxyServices.IOSSyslog, { device: this });
        iOSSystemLog.read();
    };
    IOSDevice.prototype.getInstalledApplications = function () {
        var _this = this;
        return (function () {
            return _(_this.lookupApplications()).keys().sortBy(function (identifier) { return identifier.toLowerCase(); }).value();
        }).future()();
    };
    IOSDevice.prototype.runApplication = function (applicationId) {
        var _this = this;
        return (function () {
            if (_this.$hostInfo.isWindows && (_this.$staticConfig.enableDeviceRunCommandOnWindows === null || _this.$staticConfig.enableDeviceRunCommandOnWindows === undefined)) {
                _this.$errors.fail("$%s device run command is not supported on Windows for iOS devices.", _this.$staticConfig.CLIENT_NAME.toLowerCase());
            }
            var applications = _this.lookupApplications();
            var application = applications[applicationId];
            if (!application) {
                var sortedKeys = _.sortBy(_.keys(applications));
                _this.$errors.fail("Invalid application id: %s. All available application ids are: %s%s ", applicationId, os.EOL, sortedKeys.join(os.EOL));
            }
            _this.mountImage().wait();
            var service = _this.startService(iOSProxyServices.MobileServices.DEBUG_SERVER);
            var socket = _this.$hostInfo.isWindows ? service : new net.Socket({ fd: service });
            var gdbServer = _this.$injector.resolve(iosCore.GDBServer, { socket: socket });
            var executable = util.format("%s/%s", application.Path, application.CFBundleExecutable);
            gdbServer.run([executable]);
            _this.$logger.info("Successfully run application %s on device with ID %s", applicationId, _this.getIdentifier());
        }).future()();
    };
    IOSDevice.prototype.uninstallApplication = function (applicationId) {
        var _this = this;
        return (function () {
            var afc = _this.startService(iOSProxyServices.MobileServices.INSTALLATION_PROXY);
            try {
                var result = _this.$mobileDevice.deviceUninstallApplication(afc, _this.$coreFoundation.createCFString(applicationId), null, _this.uninstallApplicationCallbackPtr);
                if (result !== 0) {
                    _this.$errors.fail("AMDeviceUninstallApplication returned '%d'.", result);
                }
            }
            finally {
            }
            _this.$logger.trace("Application %s has been uninstalled successfully.", applicationId);
        }).future()();
    };
    IOSDevice.prototype.connectToPort = function (port) {
        var interfaceType = this.getInterfaceType();
        if (interfaceType === IOSDevice.INTERFACE_USB) {
            var connectionId = this.$mobileDevice.deviceGetConnectionId(this.devicePointer);
            var socketRef = ref.alloc(CoreTypes.intType);
            this.$mobileDevice.uSBMuxConnectByPort(connectionId, this.htons(port), socketRef);
            var socketValue = socketRef.deref();
            var socket;
            if (socketValue < 0) {
                socket = new net.Socket();
                process.nextTick(function () { return socket.emit("error", new Error("USBMuxConnectByPort returned bad file descriptor")); });
            }
            else {
                socket = new net.Socket({ fd: socketValue });
                process.nextTick(function () { return socket.emit("connect"); });
            }
            return socket;
        }
        return null;
    };
    IOSDevice.prototype.htons = function (port) {
        var result = (port & 0xff00) >> 8 | (port & 0x00ff) << 8;
        return result;
    };
    IOSDevice.prototype.resolveAfc = function () {
        var service = this.$options.app ? this.startHouseArrestService(this.$options.app) : this.startService(iOSProxyServices.MobileServices.APPLE_FILE_CONNECTION);
        var afcClient = this.$injector.resolve(iOSProxyServices.AfcClient, { service: service });
        return afcClient;
    };
    IOSDevice.prototype.getFile = function (deviceFilePath) {
        var _this = this;
        return (function () {
            var afcClient = _this.resolveAfc();
            var fileToRead = afcClient.open(deviceFilePath, "r");
            var fileToWrite = _this.$options.file ? _this.$fs.createWriteStream(_this.$options.file) : process.stdout;
            var dataSizeToRead = 8192;
            var size = 0;
            while (true) {
                var data = fileToRead.read(dataSizeToRead);
                if (!data || data.length === 0) {
                    break;
                }
                fileToWrite.write(data);
                size += data.length;
            }
            fileToRead.close();
            _this.$logger.trace("%s bytes read from %s", size.toString(), deviceFilePath);
        }).future()();
    };
    IOSDevice.prototype.putFile = function (localFilePath, deviceFilePath) {
        var afcClient = this.resolveAfc();
        return afcClient.transfer(path.resolve(localFilePath), deviceFilePath);
    };
    IOSDevice.prototype.listFiles = function (devicePath) {
        var _this = this;
        return (function () {
            if (!devicePath) {
                devicePath = ".";
            }
            _this.$logger.info("Listing %s", devicePath);
            var afcClient = _this.resolveAfc();
            var walk = function (root, indent) {
                _this.$logger.info(util.format("%s %s", Array(indent).join(" "), root));
                var children = [];
                try {
                    children = afcClient.listDir(root);
                }
                catch (e) {
                    children = [];
                }
                _.each(children, function (child) {
                    walk(root + "/" + child, indent + 1);
                });
            };
            walk(devicePath, 0);
        }).future()();
    };
    IOSDevice.IMAGE_ALREADY_MOUNTED_ERROR_CODE = 3892314230;
    IOSDevice.INCOMPATIBLE_IMAGE_SIGNATURE_ERROR_CODE = 3892314163;
    IOSDevice.INTERFACE_USB = 1;
    return IOSDevice;
})();
exports.IOSDevice = IOSDevice;
$injector.register("iOSDevice", IOSDevice);
